#include <avr/interrupt.h>
#include <avr/io.h>
#include <util/delay.h>
#include <string.h>

#include "usart.h"
	
#define BUF_SIZE		32

#define S_TRIGGER 		PA0
#define S_ECHO			PD5
#define _1_US			(0.01f)
#define _1_MS			(10)

#define MIN_DISTANCE_1	20
#define MIN_DISTANCE_2	10

/* Car props */
#define BREAK 				100
#define ACCELERATION_STEP 	100
#define TURNING_STEP	 	100
#define FRICTION			30
#define USING_TRANSISTOR 	1

volatile int car_acceleration = 0;
volatile int car_turn_speed = 0;
volatile char turning_car = 0;
volatile char backwords = 0;
volatile char enable_sensors = 1;
volatile char follow = 0;

/* Timers and Sensors */
volatile unsigned int Timer1_us = 0;	/* Counts up to 1ms */
volatile unsigned int Timer1_ms = 0; 	/* Counts up to 1s */

volatile unsigned int turning = 0;
volatile unsigned int counter = 0;

unsigned long a, b = 0;
volatile int TimerOverflow = 0;
volatile int dist = 0;

/* Software PWMs */
volatile char step = 1;
volatile unsigned char MAX = 255;
volatile unsigned char PWM = 0;

volatile unsigned char DCycle1 = MAX;
volatile unsigned char DCycle2 = MAX;
volatile unsigned char DCycle3 = MAX;
volatile unsigned char DCycle4 = MAX;

volatile unsigned char PWM1_ena = 0;
volatile unsigned char PWM2_ena = 0;
volatile unsigned char PWM3_ena = 0;
volatile unsigned char PWM4_ena = 0;

/* Buffer */
char buf[BUF_SIZE];

/* Turn lights */
void turning_lights()
{
#ifdef USING_TRANSISTOR
	if (turning == 0) {
		PORTB &= ~(1 << PB7);
		PORTB &= ~(1 << PB5);
	
	} else if (turning == 1) {
		PORTB &= ~(1 << PB5);
		PORTB ^= (1 << PB7);

	} else if (turning == 2) {
		PORTB &= ~(1 << PB7);
		PORTB ^= (1 << PB5);
	
	} else if (turning == 3) {
		if (Timer1_ms == 500) {
			PORTB &= ~(1 << PB7);
			PORTB &= ~(1 << PB5);
		} else {
			PORTB |= (1 << PB7);
			PORTB |= (1 << PB5);
		}
	}
#else
	PORTB ^= (1 << PB6);
	if (turning == 0) {
		PORTB &= ~(1 << PB7);
		PORTB &= ~(1 << PB5);
	} else if (turning == 1) {
		PORTB |= (1 << PB7);
		PORTB &= ~(1 << PB5);
	} else if (turning == 2) {
		PORTB &= ~(1 << PB7);
		PORTB |= (1 << PB5);
	} else if (turning == 3) {
		PORTB |= (1 << PB7);
		PORTB |= (1 << PB5);
	}
#endif
}

/* Update custom PWM */
void PWM_update()
{
	PWM += step;

	if (PWM >= MAX) {
		step = -step;
	} else if (PWM <= 0) {
		step = -step;
	}

	/* PWM 1 */
	if (PWM == DCycle1 && PWM1_ena) {
		if (step == 1) {
			PORTD |= (1 << PD6);
		} else {
			PORTD &= ~(1 << PD6);
		}
	} else if (!PWM1_ena) {
		PORTD &= ~(1 << PD6);
	}

	/* PWM 2 */
	if (PWM == DCycle2 && PWM2_ena) {
		if (step == 1) {
			PORTD |= (1 << PD4);
		} else {
			PORTD &= ~(1 << PD4);
		}
	} else if (!PWM2_ena) {
		PORTD &= ~(1 << PD4);
	}

	/* PWM 3 */
	if (PWM == DCycle3 && PWM3_ena) {
		if (step == 1) {
			PORTB |= (1 << PB4);
		} else {
			PORTB &= ~(1 << PB4);
		}
	} else if (!PWM3_ena) {
		PORTB &= ~(1 << PB4);
	}

	/* PWM 4 */
	if (PWM == DCycle4 && PWM4_ena) {
		if (step == 1) {
			PORTB |= (1 << PB3);
		} else {
			PORTB &= ~(1 << PB3);
		}
	} else if (!PWM4_ena) {
		PORTB &= ~(1 << PB3);
	}
}

/* Custom timers and update PWM */
ISR(TIMER0_COMPA_vect)
{
	Timer1_us = (Timer1_us == _1_MS) ? 0 : Timer1_us + 1;

	if (Timer1_us == 0) {
		Timer1_ms++;
		if (Timer1_ms == 500 || Timer1_ms == 1000) {
			turning_lights();
		}

		if (Timer1_ms == 1000) {
			Timer1_ms = 0;
		}
	}
	
	PWM_update();
}

/* Timer1 detect overflows. */
ISR(TIMER1_OVF_vect)
{
	TimerOverflow++;
}

/* Turn off PWMs */
void off_pwm()
{
	if (!turning_car) {
		PWM1_ena = 0;
		PWM2_ena = 0;
		PWM3_ena = 0;
		PWM4_ena = 0;	
	}
}

/* Make the acceleration constant throughout all wheels */
void constant_accel()
{
	car_turn_speed = 0;
	turning_car = 0;

	DCycle1 = MAX - car_acceleration;
	DCycle2 = DCycle1;
	DCycle3 = DCycle1;
	DCycle4 = DCycle1;
}

/* Stop the car */
void stop()
{
	PORTD &= ~(1 << PD7);
	PORTC |= (1 << PC0);
	PORTC &= ~(1 << PC2);
	constant_accel();
	off_pwm();
}

/* Check for frontal collision */
int checkForCollision()
{
	int result = 0;
	if (backwords == 1)
		return 0;

	if (dist <= MIN_DISTANCE_2) {
		stop();
		result = 1;
	} else if (dist <= MIN_DISTANCE_1) {
		car_acceleration -= ACCELERATION_STEP + BREAK;
		if (car_acceleration <= 0) {
			car_acceleration = 0;
			off_pwm();
		}
		result = 1;
	}

	return result;
}

/* Move the car forward */
void move_forward()
{
	PORTD |= (1 << PD7);
	PORTC &= ~(1 << PC2);
	PORTC &= ~(1 << PC0);

	PWM1_ena = 0;
	PWM2_ena = 1;
	PWM3_ena = 0;
	PWM4_ena = 1;

	int res = checkForCollision();

	if (res == 0) {
		car_acceleration += ACCELERATION_STEP;
		if (car_acceleration >= MAX)
			car_acceleration = MAX;
	}

	backwords = 0;
	turning_car = 0;
	constant_accel();
}

/* Move the car backwords */
void move_backword()
{
	PORTD |= (1 << PD7);
	PORTC |= (1 << PC2);
	PORTC &= ~(1 << PC0);

	PWM1_ena = 1;
	PWM2_ena = 0;
	PWM3_ena = 1;
	PWM4_ena = 0;

	car_acceleration += ACCELERATION_STEP;
	if (car_acceleration >= MAX)
		car_acceleration = MAX;

	backwords = 1;
	turning_car = 0;
	constant_accel();
}

/* Move the car to left */
void move_left()
{
	PWM1_ena = 0;
	PWM2_ena = 1;
	PWM3_ena = 0;
	PWM4_ena = 0;
	
	car_turn_speed += TURNING_STEP;
	if (car_turn_speed >= MAX) {
		car_turn_speed = MAX;
	}

	DCycle1 = MAX - car_turn_speed;
	DCycle2 = DCycle1;
	DCycle3 = DCycle1;
	DCycle4 = DCycle1;
	turning_car = 1;
	backwords = 0;
}

/* Move the car to right */
void move_right()
{
	PWM1_ena = 0;
	PWM2_ena = 0;
	PWM3_ena = 0;
	PWM4_ena = 1;

	car_turn_speed += TURNING_STEP;
	if (car_turn_speed >= MAX) {
		car_turn_speed = MAX;
	}

	DCycle1 = MAX - car_turn_speed;
	DCycle2 = DCycle1;
	DCycle3 = DCycle1;
	DCycle4 = DCycle1;
	turning_car = 1;
	backwords = 0;
}

/* USART Receive interrupt. Get data from bluetooth. */
ISR(USART0_RX_vect)
{
	if (UCSR0A & (1 << RXC0))
	{
		char rcv = USART0_receive();

		if (rcv == 'w') {
			if (!follow)
				move_forward();

		} else if (rcv == 'n') {
			stop();
						
		} else if (rcv == 's') {
			move_backword();
			
		} else if (rcv == 'd') {			
			move_left();
		
		} else if (rcv == 'a') {
			move_right();

		} else if (rcv == 'l') {
			PORTC ^= (1 << PC1);
		
		} else if (rcv == 'q') {
			turning = 1;
		
		} else if (rcv == 'e') {
			turning = 2;
		
		} else if (rcv == 'z') {
			turning = 0;
			
		} else if (rcv == 'x') {
			turning = 3;

		} else if (rcv == 'v') {
			enable_sensors = 1 - enable_sensors;
		
		} else if (rcv == 'f') {
			follow = 1 - follow;
		}
	}
}

/* Init lights */
void initLights()
{
	DDRB |= (1 << PB5);
	DDRB |= (1 << PB6);
	DDRB |= (1 << PB7);
	
	DDRC |= (1 << PC0);
	DDRC |= (1 << PC1);
	DDRC |= (1 << PC2);
	
	PORTB &= ~(1 << PB5);
	PORTB |= (1 << PB6);
	PORTB &= ~(1 << PB7);
	
	PORTC &= ~(1 << PC0);
	PORTC &= ~(1 << PC1);
	PORTC &= ~(1 << PC2);
}

/* Init motors */
void initMotors()
{
	DDRD |= (1 << PD6);
	DDRD |= (1 << PD4);
	DDRB |= (1 << PB4);
	DDRB |= (1 << PB3);
	
	PORTD &= ~(1 << PD6);
	PORTD &= ~(1 << PD4);
	PORTB &= ~(1 << PB4);
	PORTB &= ~(1 << PB3);
}

/* Init IR sensors */
void initSensors()
{
	DDRB &= ~(1 << PB0);
	DDRB &= ~(1 << PB1);
	
	PORTB &= ~(1 << PB0);
	PORTB &= ~(1 << PB1);
}

/* Init HC-SR04 */
void initHC_SR04()
{
	DDRA |= (1 << S_TRIGGER);		/* Make trigger pin as output */

	TIMSK1 = (1 << TOIE1);
	TCCR1B |= (1 << CS10);
}

/* Init timer0 */
void initTimer()
{
	TIMSK0 |= (1 << OCIE0A);
	TCCR0A |= (1 << WGM01);
	TCCR0B |= (1 << CS01);
	OCR0A = 0xC8;
}

/* Compute distance using HC-SR04 */
int compute_dist_HC_SR04()
{
	uint32_t count = -1;

	/* Give 10us trigger pulse on trig. pin to HC-SR04 */
	TCNT1 = 0;	
	PORTA |= (1 << S_TRIGGER);
	while (TCNT1 < 100);

	TCNT1 = 0;
	PORTA |= 1 << S_TRIGGER;
	while (TCNT1 < 100);
	PORTA &= ~(1 << S_TRIGGER);

	TCNT1 = 0;
	TimerOverflow = 0;
	while (!(PIND & (1 << PIND5)));
	b = TCNT1;

	if (TimerOverflow == 0) {
		TCNT1 = 0;

		/* waiting until the echo become low again */
		while (PIND & (1 << PIND5));
		
		count = TCNT1 / 950 + (69 * TimerOverflow);
		sprintf(buf, "%ld\n", count);
		USART0_print(buf);
	} else {
		USART0_print("Out of range\n");
	}

	return count;
}

void main_loop()
{
	DDRD |= (1 << PD7);

	while(1) {
		dist = compute_dist_HC_SR04();
		int collision = checkForCollision();
		int result0 = ((PINB & (1 << PB0)) == 0);
		int result1 = ((PINB & (1 << PB1)) == 0);

		if (follow) {
			if (dist > MIN_DISTANCE_1) {
				move_forward();
			} else {
				stop();
			}
		} else {
			if (!backwords && car_acceleration > 0) {
				if (enable_sensors && !collision) {
					if (result0 && !result1)
						move_left();

					if (result1&& !result0)
						move_right();
				
					if (result0 && result1)
						stop();
				}
			}	
			
			if (car_turn_speed > 0) {
				car_turn_speed -= FRICTION;
				car_turn_speed = (car_turn_speed >= 0) ? 
					car_turn_speed : 0;
			} else {
				car_turn_speed = 1;
				turning_car = 0;
			}

			if (car_acceleration > 0) {
				car_acceleration -= FRICTION;
				car_acceleration = (car_acceleration > 0) ? 
					car_acceleration : 0;
			} else {
				car_acceleration = 0;
				off_pwm();
			} 	
		}	

		_delay_ms(50);
	}
}

/* Main function */
int main()
{
	/* Init */
	initTimer();
	initLights();
	initMotors();
	initSensors();
	initHC_SR04();
	USART0_init();
	sei();

	/* Loop */
	main_loop();
	
	while(1);

	return 0;
}
